mod api;

use std::{
    cell::RefCell,
    io::{self, Write},
    sync::{
        mpsc::{Receiver, Sender},
        Arc,
    },
};

use api::OllamaAPI;
use iced::{
    alignment::{Horizontal, Vertical},
    futures::StreamExt,
    widget::{button, column, container, row, scrollable, text, text_input},
    window, Application, Command, Length, Settings, Subscription, Theme,
};
use tokio::sync::RwLock;

use crate::api::OllamaChat;

enum UI {
    Error(String),
    Chats(UIState),
}

struct UIState {
    ollama_api: Arc<RwLock<OllamaAPI>>,
    active_chat: Option<usize>,
    chat_input: String,
    busy: bool,
    send: Sender<UIMessage>,
    recv: RefCell<Option<Receiver<UIMessage>>>,
}

impl UIState {
    fn create(
        api_url: &str,
        send: Sender<UIMessage>,
        recv: Receiver<UIMessage>,
    ) -> anyhow::Result<Self> {
        Ok(Self {
            ollama_api: Arc::new(RwLock::new(OllamaAPI::create(api_url)?)),
            active_chat: None,
            chat_input: String::from(""),
            busy: false,
            send,
            recv: RefCell::new(Some(recv)),
        })
    }
}

#[derive(Debug, Clone)]
enum UIMessage {
    Nop,
    CreateChat,
    OpenChat(usize),
    ChatInput(String),
    ChatSubmit,
    DoneGenerating,
}

fn main() -> anyhow::Result<()> {
    println!("Making request");

    let (send, recv) = std::sync::mpsc::channel();
    let mut settings = Settings::with_flags(UIFlags { send, recv });
    settings.window = window::Settings {
        size: (720, 480),
        ..window::Settings::default()
    };

    UI::run(settings)?;

    Ok(())
}

struct UIFlags {
    send: Sender<UIMessage>,
    recv: Receiver<UIMessage>,
}

impl Application for UI {
    type Executor = iced::executor::Default;
    type Message = UIMessage;
    type Theme = Theme;
    type Flags = UIFlags;

    fn new(flags: Self::Flags) -> (Self, iced::Command<Self::Message>) {
        let state: UI;
        let state_result = UIState::create("http://localhost:11434", flags.send, flags.recv);
        if let Ok(s) = state_result {
            state = UI::Chats(s);
        } else {
            state = UI::Error(String::from("Failed to initialize Ollama API"));
        }

        (state, Command::none())
    }

    fn title(&self) -> String {
        String::from("Hello World")
    }

    fn update(&mut self, message: Self::Message) -> iced::Command<Self::Message> {
        match self {
            UI::Chats(state) => match message {
                UIMessage::CreateChat => {
                    let chat = state
                        .ollama_api
                        .blocking_write()
                        .create_chat("dolphin-mixtral");
                    if let Ok(chat) = chat {
                        chat.blocking_write().send_system("Hello World!");
                    }
                    Command::none()
                }
                UIMessage::OpenChat(index) => {
                    println!("Open chat {}", index);
                    state.active_chat = Some(index);
                    Command::none()
                }
                UIMessage::ChatInput(text) => {
                    println!("Input: {}", text);
                    state.chat_input = text;
                    Command::none()
                }
                UIMessage::ChatSubmit => {
                    println!("Submit");

                    if let Some(index) = state.active_chat {
                        let content = state.chat_input.clone();
                        state.chat_input = String::from("");
                        state.busy = true;
                        let send = state.send.clone();
                        let ollama_api = state.ollama_api.clone();
                        return Command::perform(
                            async move {
                                let stream;
                                let chat;
                                {
                                    {
                                        let api = ollama_api.read().await;
                                        chat = api.chats[index].clone();
                                        let mut chat = chat.write().await;
                                        chat.send_user(&content);
                                        chat.current_message =
                                            Some(Arc::new(RwLock::new(String::from(""))));
                                    }

                                    //let mut chat = chat.write().await;
                                    stream = OllamaChat::complete(chat.clone()).await;
                                }

                                if let Ok(mut stream) = stream {
                                    while let Some(content) = stream.next().await {
                                        if content.is_err() {
                                            break;
                                        }

                                        let content = content.unwrap();

                                        let chat_arc = chat.clone();
                                        let chat = chat_arc.read().await;
                                        let msg = chat.current_message.clone();
                                        if let Some(msg) = msg {
                                            msg.write().await.push_str(&content);
                                        }

                                        print!("{}", content);
                                        let _ = io::stdout().flush();

                                        let _ = send.send(UIMessage::Nop);
                                    }
                                }
                                //state.ollama_api.complete(&x);
                            },
                            |_| UIMessage::DoneGenerating,
                        );
                    }

                    Command::none()
                }
                UIMessage::DoneGenerating => {
                    state.busy = false;
                    Command::none()
                }
                _ => Command::none(),
            },
            UI::Error(_) => Command::none(),
        }
    }

    fn view(&self) -> iced::Element<'_, Self::Message, iced::Renderer<Self::Theme>> {
        match self {
            UI::Error(message) => text(String::from("Error: ") + message)
                .width(Length::Fill)
                .height(Length::Fill)
                .vertical_alignment(Vertical::Center)
                .horizontal_alignment(Horizontal::Center)
                .into(),
            UI::Chats(state) => {
                println!("Blocking");
                let api = state.ollama_api.blocking_read();
                println!("Blocing done!");

                let mut chats: Vec<iced::Element<'_, Self::Message, iced::Renderer<Self::Theme>>> =
                    api.chats
                        .iter()
                        .enumerate()
                        .map(|c| {
                            button("amogus")
                                .on_press(UIMessage::OpenChat(c.0))
                                .width(Length::Fill)
                                .into()
                        })
                        .collect();

                chats.insert(
                    0,
                    button("Create Chat")
                        .on_press(UIMessage::CreateChat)
                        .width(Length::Fill)
                        .into(),
                );

                let chats_list = container(column(chats).padding(10.0).spacing(10))
                    .height(Length::Fill)
                    .width(Length::Fixed(300.0));

                let mut messages: Vec<
                    iced::Element<'_, Self::Message, iced::Renderer<Self::Theme>>,
                >;
                if let Some(chat_index) = state.active_chat {
                    let chat = &api.chats[chat_index].blocking_read();
                    messages = chat
                        .messages
                        .iter()
                        .map(|x| x.content.clone())
                        .map(|x| text(x).into())
                        .collect();

                    if let Some(msg) = &chat.current_message {
                        messages.push(text(msg.blocking_read()).into());
                    }
                } else {
                    messages = Vec::new();
                }

                let mut input_box = text_input("Your message", &state.chat_input);
                if !state.busy {
                    input_box = input_box
                        .on_input(|x| UIMessage::ChatInput(x))
                        .on_submit(UIMessage::ChatSubmit)
                }

                let active_chat = container(
                    column![
                        scrollable(container(column(messages).spacing(10)))
                            .width(Length::Fill)
                            .height(Length::Fill),
                        input_box
                    ]
                    .padding(10)
                    .spacing(10),
                )
                .width(Length::Fill)
                .height(Length::Fill);

                container(row![chats_list, active_chat]).into()

                /*scrollable(
                    container(column![content, amogus].spacing(20).max_width(480))
                        .padding(40)
                        .center_x(),
                )
                .into()*/
            }
        }
    }

    fn theme(&self) -> Self::Theme {
        Theme::Dark
    }

    fn subscription(&self) -> iced::Subscription<Self::Message> {
        match self {
            UI::Chats(state) => iced::subscription::unfold(
                "external messages",
                state.recv.take(),
                move |recv| async move {
                    let msg = recv.as_ref().unwrap().recv();

                    if let Err(_) = msg {
                        return (UIMessage::Nop, recv);
                    }

                    let msg = msg.unwrap();
                    (msg, recv)
                },
            ),
            _ => Subscription::none(),
        }
    }
}
